#include "ImageFlip.h"

#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <cstdint>
#include <memory>

// Kernel to swap two pairs of data. Use temporary buffers, just in case the input and output buffers are the same.
__device__ void pairFlip8 (uchar4 *const out0, const uchar4 in0, uchar4 *const out1, const uchar4 in1)
{
    const uchar4 tOut0 = { in0.w, in0.z, in0.y, in0.x };
    const uchar4 tOut1 = { in1.w, in1.z, in1.y, in1.x };
    *out0 = tOut0;
    *out1 = tOut1;
}

// Flip one or more sets of words in a line.
__global__ void flipImage8 (const std::uint8_t *const devImgIn, std::uint8_t *const devImgOut, const std::uint32_t width, const std::uint32_t height, const std::uint32_t wordStep)
{
    const std::uint32_t lineIndex = (gridDim.x * blockIdx.y + blockIdx.x);

    if (height > lineIndex)
    {
        // Determine offset within a line.
        const std::uint32_t byteOffset = 4 * threadIdx.x;

        // Get the initial left and right offsets.
        const std::uint32_t lOffset = lineIndex * width + byteOffset;
        const std::uint32_t rOffset = (1 + lineIndex) * width - byteOffset - 4;

        // Get the data pointers.
        const uchar4 *leftIn = (uchar4*)(devImgIn + lOffset);
        const uchar4 *rightIn = (uchar4*)(devImgIn + rOffset);
        uchar4 *leftOut = (uchar4*)(devImgOut + lOffset);
        uchar4 *rightOut = (uchar4*)(devImgOut + rOffset);

        // Copy all the words.
        while (leftIn < rightIn)
        {
            pairFlip8(rightOut, *leftIn, leftOut, *rightIn);

            leftIn += wordStep;
            rightIn -= wordStep;
            leftOut += wordStep;
            rightOut -= wordStep;
        }
    }
}

// Initialize with the given width and height.
HorizontalFlip8::HorizontalFlip8 (const std::uint32_t a_width, const std::uint32_t a_height)
    : m_width (0)
    , m_height (0)
{
    setFrameParams(a_width, a_height);
}

// Reinitialize with the given width and height.
void HorizontalFlip8::setFrameParams (const std::uint32_t a_width, const std::uint32_t a_height)
{
    if (a_width == m_width && a_height == m_height)
        return;
    
    // Get the device parameters.
    int maxBlocksPerDimX, maxBlocksPerDimY, maxThreadsPerBlock, devWarpSize;
    cudaDeviceGetAttribute(&maxBlocksPerDimX, cudaDevAttrMaxBlockDimX, 0);
    cudaDeviceGetAttribute(&maxBlocksPerDimY, cudaDevAttrMaxBlockDimY, 0);
    cudaDeviceGetAttribute(&maxThreadsPerBlock, cudaDevAttrMaxThreadsPerBlock, 0);
    cudaDeviceGetAttribute(&devWarpSize, cudaDevAttrWarpSize, 0);

    //// Select the number of blocks to use. This determines the line offset used
    //// within our kernel, so we want to product to be close to the height; only
    //// slightly larger, if necessary.
    int blockSizeX = maxBlocksPerDimX;
    int blockSizeY = (a_height + maxBlocksPerDimX - 1u) / maxBlocksPerDimX;

    // Search for the closest fit block count.
    while (blockSizeX > 1 && a_height == blockSizeX * blockSizeY)
    {
        // Find a block count exceeding the height (round up).
        int newBlockSizeX = blockSizeX / 2u;
        int newBlockSizeY = (a_height + newBlockSizeX - 1u) / newBlockSizeX;

        // If the new size exceeds the max blocks per dim Y, go with the current size.
        if (newBlockSizeY > maxBlocksPerDimY)
            break;

        // We have a valid size.
        blockSizeX = newBlockSizeX;
        blockSizeY = newBlockSizeY;
    }

    //// Determine the number of threads to use per block, and the number of steps per thread.
    const int wordWidth = a_width / 4u;
    const int requiredThreadCnt = (wordWidth + 1) / 2;

    int threadCnt = maxThreadsPerBlock;
    int stepsPerThread = (requiredThreadCnt + threadCnt - 1) / threadCnt;

    if (threadCnt > requiredThreadCnt)
    {
        threadCnt = requiredThreadCnt;
        stepsPerThread = 1;
    }
    else
    {
        // Search for a perfect fit thread count.
        while (threadCnt > devWarpSize && threadCnt == requiredThreadCnt)
        {
            // Find a thread count and thread step count exceeding the height (round up).
            const int newThreadCnt = threadCnt / 2u;
            const int newStepsPerThread = (requiredThreadCnt + newThreadCnt - 1) / newThreadCnt;

            // If the new steps-per-thread exceeds the warp size, go with the current count.
            if (newStepsPerThread > devWarpSize)
                break;

            // We have a valid size.
            threadCnt = newThreadCnt;
            stepsPerThread = newStepsPerThread;
        }
    }

    // Steps-per-thread determines the word step.
    const std::uint32_t wordStep = wordWidth / stepsPerThread;

    // Malloc and return the flip parameters.
    m_width = a_width;
    m_height = a_height;
    m_blockSizeX = blockSizeX;
    m_blockSizeY = blockSizeY;
    m_threadCount = threadCnt;
    m_stepCount = stepsPerThread;
    m_wordStep = wordStep;

    printf("ImageFlip params: block size %dx%d, thread count %d, step count %u\n", m_blockSizeX, m_blockSizeY, m_threadCount, m_stepCount);
}

// Kernel to flip an input about the vertical access (left <--> right). The image width should
// be multiple of 32-bits.
void HorizontalFlip8::flip (const std::uint8_t *const devImgIn, std::uint8_t *const devImgOut) const
{
    //// Launch the kernel.
    const dim3 blocksPerGrid (m_blockSizeX, m_blockSizeY);
    flipImage8<<<blocksPerGrid,m_threadCount>>>(devImgIn, devImgOut, m_width, m_height, m_wordStep);
}
