/* FILE:        SimpleDVP.cpp
 * DATE:        5/7/2014
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2014, BitFlow, Inc.
 * DESCRIPTION: A simple demonstration of the BFDVP API. This example will
 *              acquire from a BitFlow frame grabber, DMA each acquired buffer
 *              to the zero index CUDA device, then copy the buffer into an
 *              OpenGL buffer for display in a FreeGLUT window.
 */

#define NOMINMAX

// Local helper includes.
#include "BFDVPsimpleGLUT.h"
#include "SimpleTimers.hpp"

// CUDA processing methods.
#include "ImageFlip.h"

// Standard includes.
#include <iostream>

#include <atomic>
#include <thread>
#include <mutex>

#include <vector>
#include <string>

// BitFlow includes.
#include <BFApi.h>
#include <BFErApi.h>
#include <BiApi.h>

// BFDVP includes.
#include <BFDVPApi.h>

// CUDA includes.
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <cuda_profiler_api.h>

// CUDA profiler definitions.
#define CUDA_PROFILER_ENABLED 1

#if CUDA_PROFILER_ENABLED
#   define CUDA_PROFILER_START() (cudaProfilerStart())
#   define CUDA_PROFILER_STOP()  (cudaProfilerStop())
#else
#   define CUDA_PROFILER_START() (CUDA_SUCCESS)
#   define CUDA_PROFILER_STOP()  (CUDA_SUCCESS)
#endif

// The number of system-device buffer pairs to allocate for acquisition and
// processing. There must be at least two acquisition buffer pairs, plus an
// extra processing buffer pair, for a minimum valid count of 3 buffers.
BFU32           vfgNumber           = ~BFU32(0);
BFU32           bufferPairCount     = 25;

BFU32           imageWidth          = 0;
BFU32           imageHeight         = 0;
BFU32           imageBitDepth       = 0;
BFU32           imageByteSize       = 0;

bool            imageFlipEnabled    = false;
bool            useDvpDma           = true;
bool            outToHost           = false;
bool            startInFullScreen   = false;
bool            showOverlay         = true;
bool            insanityModeEnabled = false;

RateAverager<>  processingRateAverager;
std::mutex      processingRateAveragerMutex;

// Do whatever processing we want to for the acquired buffer.
std::atomic<bool>             exitThreadsNow   (false);
std::atomic<PBFDVPSysDevPair> processBufferIn  (nullptr);
std::atomic<PBFDVPSysDevPair> processBufferOut (nullptr);

void process_buffers (void)
{
    BFRC rc;

    const HorizontalFlip8 flipKernel (imageWidth, imageHeight);

    while (!exitThreadsNow.load())
    {
        std::this_thread::yield();

        // Retrieve the last acquired frame and the current output buffer. For the purposes
        // of this example, don't worry whether or not we have previously processed this image.
        PBFDVPSysDevPair bufPairIn = processBufferIn.load();
        PBFDVPSysDevPair bufPairOut = processBufferOut.load();
        if (!bufPairIn || !bufPairOut)
            continue;
        
        if (useDvpDma)
        {
            // DMA the buffer to the device (necessary, to do any CUDA processing).
            rc = BFDVPSys2Dev(bufPairIn, nullptr);
            if (BF_OK != rc)
            {
                std::cerr << "Sys2Dev transfer failed.\n";
                continue;
            }
        }
        else
        {
            // Attempt to copy the system buffer to the device input buffer.
            if (CUDA_SUCCESS != cuMemcpyHtoD(bufPairIn->devBuffer, bufPairIn->sysBuffer, imageByteSize))
            {
                std::cerr << "cuMemcpyHtoD transfer failed.\n";
                continue;
            }
        }

        // Do something with the GPU's copy of the buffer.
        if (imageFlipEnabled && 8 == imageBitDepth)
        {
            // Flip the input buffer image horizontally, storing results in the output buffer.
            flipKernel.flip((std::uint8_t*)bufPairIn->devBuffer, (std::uint8_t*)bufPairOut->devBuffer);

            // Pass the processed buffer to the display.
            setMainBFDVPsimpleGLUT(bufPairOut->devBuffer);

            if (outToHost)
            {
                if (useDvpDma)
                {
                    // DMA the processed buffer to the system.
                    rc = BFDVPDev2Sys(bufPairOut, nullptr);
                    if (BF_OK != rc)
                    {
                        std::cerr << "Dev2Sys transfer failed.\n";
                        continue;
                    }
                }
                else
                {
                    // Attempt to copy the to the device output buffer to the system buffer.
                    if (CUDA_SUCCESS != cuMemcpyDtoH(bufPairOut->sysBuffer, bufPairOut->devBuffer, imageByteSize))
                    {
                        std::cerr << "cuMemcpyDtoH transfer failed.\n";
                        continue;
                    }
                }
            }
        }
        else
        {
            // Pass the buffer for display.
            setMainBFDVPsimpleGLUT(bufPairIn->devBuffer);
        }

        // Iterate the processing rate averager.
        {
            const std::lock_guard<std::mutex> processingRateAveragerLock (processingRateAveragerMutex);
            processingRateAverager.count();
        }

        // Normally we want to give things a break, so the display doesn't choke.
        if (!insanityModeEnabled)
            cudaDeviceSynchronize();
    }
}

// The acquisition thread loop.
void enterAcquisitionMainLoop ( const Bd hBoard
                              , const PBIBA pBufArray
                              , PBFDVPSysDevPair sysDevPairs )
{
    BFRC rc;

    // Loop acquisition until the user presses a key, or the acquisition aborts.
    BFU32 timeoutCnt = 0;
    BiCirHandle cirHandle;

    // Setup counters.
    {
        const std::lock_guard<std::mutex> processingRateAveragerLock (processingRateAveragerMutex);
        processingRateAverager.reset(2 * bufferPairCount);
    }
    RateAverager<> acquisitionRateAverager (2 * bufferPairCount);
    PassiveTimer<> dispUpdateTimer (1000 / 10);

    // Loop indefinitely, processing every frame as it becomes available.
    while (!exitThreadsNow.load())
    {
        // Print a status message at about 60Hz (or less).
        if (dispUpdateTimer.timeout())
        {
            size_t procdFrmCnt;
            double procdRteAvg;
            {
                const std::lock_guard<std::mutex> processingRateAveragerLock (processingRateAveragerMutex);
                procdFrmCnt = processingRateAverager.total_periods();
                procdRteAvg = processingRateAverager.average_rate();
            }
            
            const size_t acqdFrmCnt = acquisitionRateAverager.total_periods();
            const double acqdRteAvg = acquisitionRateAverager.average_rate();

            std::cout << acqdFrmCnt << " acquired, " << procdFrmCnt << " processed, " << timeoutCnt << " timeouts\r";

            if (showOverlay)
            {
                char overlayText [256];
                sprintf(overlayText, "Disp. Rate: %6.2f FPS\nProc. Rate: %6.2f FPS\nAcq.  Rate: %6.2f FPS\nTxfr. Mode: %s"
                    , getBFDVPsimpleGLUTdisplayRateAverage()
                    , procdRteAvg
                    , acqdRteAvg
                    , useDvpDma ? "DVP" : "CUDA");

                setBFDVPsimpleGLUToverlayText(overlayText);
            }
            else
                setBFDVPsimpleGLUToverlayText("");
        }

        // Wait for a frame.
        rc = BiCirWaitDoneFrame(hBoard, pBufArray, 100, &cirHandle);
        switch (rc)
        {
        case BF_OK:
            processBufferIn.store( &sysDevPairs[cirHandle.BufferNumber] );
            processBufferOut.store( &sysDevPairs[bufferPairCount - 1] );

            // Make the buffer available to the acquisition engine again.
            acquisitionRateAverager.count();
            BiCirStatusSet(hBoard, pBufArray, cirHandle, BIAVAILABLE);
            continue;

        case BI_ERROR_CIR_WAIT_TIMEOUT:
            timeoutCnt++;
            continue;
        }

        // At any other error, quit acquisition. Abort indicates a successful exit.
        if (BI_CIR_ABORTED != rc)
            BiErrorShow(hBoard, rc);
        break;
    }
}

// External key-press handler for the BFDVPsimpleGLUT display.
bool keyPressHandler (const char key)
{
    switch (key)
    {
    case 'o': case 'O':
        showOverlay = !showOverlay;
        std::cout << "\n" << (showOverlay ? "Enabled" : "Disabled") << " the text overlay.\n";
        return true;

    case 'd': case 'D':
        useDvpDma = !useDvpDma;
        std::cout << "\n" << (useDvpDma ? "Enabled" : "Disabled") << " DVP DMA.\n";
        return true;

    case '\t':
        if (8 != imageBitDepth)
        {
            imageFlipEnabled = false;
            std::cerr << "\nImage Flip only supported for 8-bit images.\n";
        }
        else if (imageFlipEnabled = !imageFlipEnabled)
            std::cout << "\nImage Flip ENABLED\n";
        else
            std::cout << "\nImage Flip DISABLED\n";
        return true;

    case 'h': case 'H':
        if (outToHost = !outToHost)
            std::cout << "\nOutput to Host transfer ENABLED\n";
        else
            std::cout << "\nOutput to Host transfer DISABLED\n";
        return true;

    case 'i': case 'I':
        if (insanityModeEnabled = !insanityModeEnabled)
            std::cout << "\nInsanity Mode ENABLED\n";
        else
            std::cout << "\nInsanity Mode DISABLED\n";
        return true;
    }

    return false;
}

typedef std::vector<std::string> ArgsVector;
bool processArgs (ArgsVector const& args)
{
    std::string badArgument;

    for (auto &arg : args)
    {
        if (1 == sscanf(arg.c_str(), "--board=%u", &vfgNumber))
            std::cout << "Selected BitFlow board index " << vfgNumber << ".\n";
        else if (1 == sscanf(arg.c_str(), "--acqBufs=%u", &bufferPairCount))
        {
            std::cout << "Selected " << bufferPairCount << " acquisition buffers.\n";
            if (3 > bufferPairCount)
            {
                std::cerr << "ERROR: At least two buffers are required for acquisition, plus one for output!\n";
                badArgument = arg;
                break;
            }
        }
        else if (arg == "--flip")
        {
            std::cout << "Enabled Image Flip mode.\n";
            imageFlipEnabled = true;
        }
        else if (arg == "--noDvp")
        {
            std::cout << "Disabled DVP DMA.\n";
            useDvpDma = false;
        }
        else if (arg == "--out2host")
        {
            std::cout << "Enabled Output to Host transfer.\n";
            outToHost = true;
        }
        else if (arg == "--fullscreen")
        {
            std::cout << "The application will start in full-screen mode.\n";
            startInFullScreen = true;
        }
        else if (arg == "--noOverlay")
        {
            std::cout << "Disabled the overlay.\n";
            showOverlay = false;
        }
        else if (arg == "--insanity")
        {
            std::cout << "Enabled Insanity Mode.\n";
            insanityModeEnabled = true;
        }
        else
        {
            std::cerr << "Unknown argument!\n";
            badArgument = arg;
            break;
        }
    }

    if (badArgument.size())
    {
        if (badArgument != "?" && badArgument != "/?" && badArgument != "h" && badArgument != "-h" && badArgument != "--help" && badArgument != "help")
            std::cerr << "ERROR: Found a bad input argument: " << badArgument << "\n";

        /*   cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc */
        std::cerr <<
            "Known arguments are: \n"
            "  --board=N       : Set the BitFlow board to be opened (asks, by default).\n"
            "  --acqBufs=N     : The Number of acquisition buffers (default 10, minimum 2).\n"
            "  --flip          : Enable Image Flip at startup.\n"
            "  --noDvp         : Disable DVP DMA at startup (i.e., use CUDA memcpy instead).\n"
            "  --out2host      : Enable Output to Host transfer of the result buffer.\n"
            "  --fullscreen    : Set the application window to fullscreen mode.\n"
            "  --noOverlay     : Disable the text overlay.\n"
            "  --insanity      : Enable Insanity Mode at startup.\n";
        return false;
    }

    return true;
}

// The main method.
int main (const int argc, char *argv[])
{
    BFRC rc;

    // Process any arguments.
    const ArgsVector args (argv + 1, argv + argc);
    if (!processArgs(args))
        return EXIT_FAILURE;
    
    // Open the board.
    Bd hBoard;
    if ( !(~vfgNumber) )
    {
        // Query the user for a board to open.
        BFU32 BoardType, BoardNum;
        BFU32 Init, SerNum;
        if ( IDCANCEL == DoBrdOpenDialog(BOD_HIDEJUSTOPEN, FF_BITFLOW_MODERN, &BoardType, &BoardNum, &Init, &SerNum) )
            return EXIT_SUCCESS;
    
        rc = BiBrdOpen(BoardType, BoardNum, &hBoard);
        if (BF_OK != rc)
        {
            std::cerr << "Failed to open BitFlow board " << BoardNum << ".\n";
            return EXIT_FAILURE;
        }
    }
    else
    {
        rc = BiBrdOpen(BiTypeAny, vfgNumber, &hBoard);
        if (BF_OK != rc)
        {
            std::cerr << "Failed to open BitFlow board " << vfgNumber << ".\n";
            return EXIT_FAILURE;
        }
    }

    // Retrieve the requisite board parameters.
    BiBrdInquire(hBoard, BiCamInqXSize, &imageWidth);
    BiBrdInquire(hBoard, BiCamInqYSize0, &imageHeight);
    BiBrdInquire(hBoard, CiCamInqPixBitDepth, &imageBitDepth);

    // Verify that we support the input bitdepth.
    switch (imageBitDepth)
    {
    case 8: case 10: case 12: case 14: case 16:
        break;

    default:
        // Unwind
        BiBrdClose(hBoard);

        std::cerr << "Bit-depth " << imageBitDepth << " unsupported by this demo.\n";
        return EXIT_FAILURE;
    }
    
    const BFU32 imageByteDepth = (imageBitDepth + 7) / 8;
    imageByteSize = imageWidth * imageHeight * imageByteDepth;
    
    // At application exit, call cudaDeviceReset.
    atexit([]() { cudaDeviceReset(); });
    
    // Setup the CUDA context.
    if (CUDA_SUCCESS != CUDA_PROFILER_START())
    {
        BiBrdClose(hBoard);

        std::cerr << "Unable to start the CUDA profiler.\n";
        return EXIT_FAILURE;
    }

    if (CUDA_SUCCESS != cuInit(0))
    {
        // Unwind
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Unable to initialize the CUDA library.\n";
        return EXIT_FAILURE;
    }

    CUcontext cudaCtx;
    if (CUDA_SUCCESS != cuCtxCreate(&cudaCtx, 0, 0))
    {
        // Unwind
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to create the CUDA context object.";
        return EXIT_FAILURE;
    }

    if (CUDA_SUCCESS != cuCtxSetCurrent(cudaCtx))
    {
        // Unwind
        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to set the current CUDA context.";
        return EXIT_FAILURE;
    }

    // Initialize the DVP context.
    rc = BFDVPInit(BFDVPInit_Default);
    if (BF_OK != rc)
    {
        // Unwind
        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to initialize the DVP context.\n";
        return EXIT_FAILURE;
    }

    // Allocate system and GPU device buffers using the BFDVPAllocAndBindPairs method.
    std::vector<BFDVPSysBuf>        systemBuffers   (bufferPairCount);
    std::vector<BFDVPSysDevPair>    sysDevPairs     (bufferPairCount);
    rc = BFDVPAllocAndBindPairs(systemBuffers.data(), sysDevPairs.data(), imageWidth, imageHeight, imageBitDepth, bufferPairCount, nullptr);
    if (BF_OK != rc)
    {
        // Unwind
        BFDVPDeinit();

        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to allocate and bind the system and device buffer pairs.\n";
        return EXIT_FAILURE;
    }

    // Assign the allocated system buffers to the BitFlow acquisition engine. The final
    // buffer is reserved for processing, and therefore excluded from the acquisition engine.
    BIBA bufArray;
    rc = BiBufferAssign(hBoard, &bufArray, systemBuffers.data(), bufferPairCount - 1);
    if (BF_OK != rc)
    {
        // Unwind
        BFDVPBreakAndFreePairs(sysDevPairs.data(), bufferPairCount);
        BFDVPDeinit();

        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to assign the system buffers to the BitFlow acquisition engine.\n";
        return EXIT_FAILURE;
    }

    // Configure a circular acquisition.
    rc = BiCircAqSetup(hBoard, &bufArray, CirErStop, BiAqEngJ);
    if (BF_OK != rc)
    {
        // Unwind
        BiBufferUnassign(hBoard, &bufArray);
        BFDVPBreakAndFreePairs(sysDevPairs.data(), bufferPairCount);
        BFDVPDeinit();

        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to setup the frame grabber for a circular acquisition.\n";
        return EXIT_FAILURE;
    }
    
    std::cout <<
    /*   cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc */
        "Keyboard controls:\n"
        "  [Q]|ESC       : Exit the program.\n"
        "  [+/-]         : Zoom in/out.\n"
        "  [1]           : Zoom to 100%, and reset panning.\n"
        "  [F]           : Zoom to fit the display window, and reset panning.\n"
        "  TAB           : Enable/disable Image Flip (8-bit images only).\n"
        "  [D]           : Toggle between DVP and plain CUDA DMA.\n"
        "  [H]           : Enable/disable Output to Host transfer of the result buffer.\n"
        "  [O]           : Toggle visibility of the text Overlay.\n"
        "  [F11]         : Toggle fullscreen display.\n"
        "  [I]           : Enable/disable Insanity Mode for unmetered processing (may\n"
        "                  reduce display rate).\n"
        "Mouse controls:\n"
        "  Scroll-wheel  : Zoom in or out.\n"
        "  Right-click   : Press, hold and drag, to pan the display.\n\n"
        "Closing the display window exits.\n\n";
    
    std::thread glutThread, acqThread;

    // Start acquisition.
    rc = BiCirControl(hBoard, &bufArray, BISTART, BiAsync);
    if (BF_OK == rc)
    {
        // Setup the display key-press handler.
        setBFDVPsimpleGLUTkeyCallback(&keyPressHandler);

        // Enter the BFDVPsimpleGLUT main loop.
        glutThread = std::thread( [&](void)
        {
            enterBFDVPsimpleGLUT(argc, argv, imageWidth, imageHeight, imageBitDepth, startInFullScreen);

            // Abort acquisition.
            exitThreadsNow.store(true);
            BiCirControl(hBoard, &bufArray, BIABORT, BiAsync);
        } );

        // Enter the acquisition main loop.
        acqThread = std::thread( &enterAcquisitionMainLoop, hBoard, &bufArray, sysDevPairs.data() );

        // Enter the processing loop. CUDA operations need to take place in _this_ thread.
        process_buffers();
    }

    // Cleanup everything.
    if (glutThread.joinable())
        glutThread.join();
    if (acqThread.joinable())
        acqThread.join();

    BiCircCleanUp(hBoard, &bufArray);
    BiBufferUnassign(hBoard, &bufArray);
    BFDVPBreakAndFreePairs(sysDevPairs.data(), bufferPairCount);
    BFDVPDeinit();

    cuCtxDestroy(cudaCtx);
    CUDA_PROFILER_STOP();
    BiBrdClose(hBoard);

    // All done!
    return EXIT_SUCCESS;
}
