/* FILE:        SimpleDVP.cpp
 * DATE:        2/23/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2014, BitFlow, Inc.
 * DESCRIPTION: A simple demonstration of the BFDVP API. This example will
 *              acquire a sequence of sub-images from a BitFlow frame grabber,
 *              and DMA each acquired sub-image buffer to a single whole-image
 *              buffer on the zero index CUDA device.
 */

// Standard includes.
#include <iostream>

#include <atomic>
#include <thread>
#include <mutex>

#include <vector>
#include <string>

// BitFlow includes.
#include <BFApi.h>
#include <BFErApi.h>
#include <BiApi.h>

// BFDVP includes.
#include <BFDVPApi.h>

// CUDA includes.
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <cuda_profiler_api.h>

// CUDA profiler definitions.
#define CUDA_PROFILER_ENABLED 1

#if CUDA_PROFILER_ENABLED
#   define CUDA_PROFILER_START() (cudaProfilerStart())
#   define CUDA_PROFILER_STOP()  (cudaProfilerStop())
#else
#   define CUDA_PROFILER_START() (CUDA_SUCCESS)
#   define CUDA_PROFILER_STOP()  (CUDA_SUCCESS)
#endif

int main (int argc, char *argv[])
{
    BFRC rc;
    
    // The number of acquisition buffers per each whole image.
    const BFU32 subImageCnt = 10;

    // Query the user for a board to open.
    BFU32 BoardType, BoardNum;
    BFU32 Init, SerNum;
    if ( IDCANCEL == DoBrdOpenDialog(BOD_HIDEJUSTOPEN, FF_BITFLOW_MODERN, &BoardType, &BoardNum, &Init, &SerNum) )
        return EXIT_SUCCESS;
    
    // Open the board.
    Bd hBoard;
    rc = BiBrdOpen(BoardType, BoardNum, &hBoard);
    if (BF_OK != rc)
    {
        std::cerr << "Failed to open BitFlow board " << BoardNum << ".\n";
        return EXIT_FAILURE;
    }

    // Retrieve the requisite board parameters.
    BFU32 imageWidth, subImageHeight, imageBitDepth;
    BiBrdInquire(hBoard, BiCamInqXSize, &imageWidth);
    BiBrdInquire(hBoard, BiCamInqYSize0, &subImageHeight);
    BiBrdInquire(hBoard, CiCamInqPixBitDepth, &imageBitDepth);

    // Verify that we support the input bitdepth.
    switch (imageBitDepth)
    {
    case 8: case 10: case 12: case 14: case 16:
        break;

    default:
        // Unwind
        BiBrdClose(hBoard);

        std::cerr << "Bit-depth " << imageBitDepth << " unsupported by this demo.\n";
        return EXIT_FAILURE;
    }
    
    // Calculate common parameters.
    const BFU32 imageByteDepth = (imageBitDepth + 7) / 8;

    const BFU32 subImageByteSize = imageWidth * subImageHeight * imageByteDepth;

    const BFU32 wholeImageHeight = subImageHeight * subImageCnt;
    const BFU32 wholeImageByteSize = imageWidth * wholeImageHeight * imageByteDepth;
    
    // At application exit, call cudaDeviceReset.
    atexit([]() { cudaDeviceReset(); });
    
    // Setup the CUDA context.
    if (CUDA_SUCCESS != CUDA_PROFILER_START())
    {
        BiBrdClose(hBoard);

        std::cerr << "Unable to start the CUDA profiler.\n";
        return EXIT_FAILURE;
    }

    if (CUDA_SUCCESS != cuInit(0))
    {
        // Unwind
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Unable to initialize the CUDA library.\n";
        return EXIT_FAILURE;
    }

    CUcontext cudaCtx;
    if (CUDA_SUCCESS != cuCtxCreate(&cudaCtx, 0, 0))
    {
        // Unwind
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to create the CUDA context object.";
        return EXIT_FAILURE;
    }

    if (CUDA_SUCCESS != cuCtxSetCurrent(cudaCtx))
    {
        // Unwind
        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to set the current CUDA context.";
        return EXIT_FAILURE;
    }

    // Initialize the DVP context.
    rc = BFDVPInit(BFDVPInit_Default);
    if (BF_OK != rc)
    {
        // Unwind
        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to initialize the DVP context.\n";
        return EXIT_FAILURE;
    }

    // Allocate system buffers.
    std::vector<BFDVPSysBuf> systemBuffers (subImageCnt);
    rc = BFDVPSysBuffersAlloc(systemBuffers.data(), imageWidth, subImageHeight, imageBitDepth, subImageCnt, nullptr);
    if (BF_OK != rc)
    {
        // Unwind
        BFDVPDeinit();

        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to allocate the system buffer.\n";
        return EXIT_FAILURE;
    }

    // Allocate a single large CUDA buffer.
    CUdeviceptr cudaBuffer;
    if (CUDA_SUCCESS != cuMemAlloc(&cudaBuffer, wholeImageByteSize))
    {
        // Unwind
        BFDVPSysBuffersFree(systemBuffers.data(), subImageCnt);
        BFDVPDeinit();

        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to allocate the CUDA buffer.\n";
        return EXIT_FAILURE;
    }

    // Bind the image sub-buffers.
    std::vector<BFDVPSysDevPair> sysDevPairs (subImageCnt);

    for (BFU32 i = 0; i < subImageCnt; i++)
    {
        const CUdeviceptr cudaSubBuffer = cudaBuffer + subImageByteSize * i;

        rc = BFDVPBindPair(&sysDevPairs[i], systemBuffers[i], cudaSubBuffer, imageWidth, subImageHeight, imageBitDepth);
        if (BF_OK != rc)
        {
            // Unwind.
            while (i > 0)
            {
                --i;
                BFDVPBreakPair(&sysDevPairs[i]);
            }

            cuMemFree(cudaBuffer);
            
            BFDVPSysBuffersFree(systemBuffers.data(), subImageCnt);
            BFDVPDeinit();

            cuCtxDestroy(cudaCtx);
            CUDA_PROFILER_STOP();
            BiBrdClose(hBoard);

            std::cerr << "Failed to bind the system and CUDA buffers.\n";
            return EXIT_FAILURE;
        }
    }

    // Assign the allocated system buffers to the BitFlow acquisition engine.
    BIBA bufArray;
    rc = BiBufferAssign(hBoard, &bufArray, systemBuffers.data(), subImageCnt);
    if (BF_OK != rc)
    {
        // Unwind
        for (BFU32 i = 0; i < subImageCnt; i++)
            BFDVPBreakPair(&sysDevPairs[i]);

        cuMemFree(cudaBuffer);
            
        BFDVPSysBuffersFree(systemBuffers.data(), subImageCnt);
        BFDVPDeinit();

        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to assign the system acquisition buffers.\n";
        return EXIT_FAILURE;
    }

    // Configure a circular acquisition.
    rc = BiCircAqSetup(hBoard, &bufArray, CirErStop, BiAqEngJ);
    if (BF_OK != rc)
    {
        // Unwind
        BiBufferUnassign(hBoard, &bufArray);
        
        for (BFU32 i = 0; i < subImageCnt; i++)
            BFDVPBreakPair(&sysDevPairs[i]);

        cuMemFree(cudaBuffer);
            
        BFDVPSysBuffersFree(systemBuffers.data(), subImageCnt);
        BFDVPDeinit();

        cuCtxDestroy(cudaCtx);
        CUDA_PROFILER_STOP();
        BiBrdClose(hBoard);

        std::cerr << "Failed to setup the frame grabber for a circular acquisition.\n";
        return EXIT_FAILURE;
    }

    // Start the circular acquisition, and loop until the user presses a key.

    std::cout << "\nEntering the acquisition main loop. Press any key to exit.\n\n";

    BiCirControl(hBoard, &bufArray, BISTART, BiAsync);

    BFU32 wholeImageTransferCnt = 0;
    while (!BFkbhit())
    {
        // Wait for a frame.
        BiCirHandle cirHandle;
        rc = BiCirWaitDoneFrame(hBoard, &bufArray, INFINITE, &cirHandle);
        if (BF_OK != rc)
        {
            std::cerr << "Received an acquisition error. Aborting.\n";
            break;
        }

        // Transfer the sub-image from the host to the device.
        rc = BFDVPSys2Dev(&sysDevPairs[cirHandle.BufferNumber], nullptr);
        if (BF_OK != rc)
        {
            std::cerr << "Recived a DVP transfer error. Aborting.\n";
            break;
        }

        // Make the sub-image available to the acquisition engine.
        BiCirStatusSet(hBoard, &bufArray, cirHandle, BIAVAILABLE);

        // Record whole image transfers.
        if ((cirHandle.FrameCount - 1) % subImageCnt == 0)
            wholeImageTransferCnt++;

        // Print a message.
        printf("\rSub-Image Cnt: %u\tWhole Image Cnt: %u", cirHandle.FrameCount, wholeImageTransferCnt);
    }

    BiCirControl(hBoard, &bufArray, BISTOP, BiWait);
    
    // Cleanup.
    BiCircCleanUp(hBoard, &bufArray);
    BiBufferUnassign(hBoard, &bufArray);
    
    for (BFU32 i = 0; i < subImageCnt; i++)
        BFDVPBreakPair(&sysDevPairs[i]);

    cuMemFree(cudaBuffer);
            
    BFDVPSysBuffersFree(systemBuffers.data(), subImageCnt);
    BFDVPDeinit();

    cuCtxDestroy(cudaCtx);
    CUDA_PROFILER_STOP();
    BiBrdClose(hBoard);


    // All done!
    return EXIT_SUCCESS;
}
