#ifndef INCLUDED__B_F__D_V_P__API__H
#define INCLUDED__B_F__D_V_P__API__H

/* FILE:        BFDVPApi.h
 * DATE:        4/29/2014
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2014, BitFlow, Inc.
 * DESCRIPTION: BitFlow API to interface with an NVIDIA GPU supporting the
 *              GPUDirect for Video protocol.
 */

#ifdef __cplusplus
extern "C" {
#endif

#include "BFDVPDef.h"
#include "BFDVPError.h"

/* BFDVPInit -- Initialize the DVP context for the given flags. Must be called once, prior to using any other BFDVP methods.
 *  flags           Flags indicating how the DVP context should be initialized.
 */
BFDLL BFRC BFCAPI BFDVPInit (BFDVPInitFlags flags);

/* BFDVPDeinit -- Deinitialize the DVP context. Should be called once at cleanup, given a prior successful call to BFDVPInit.
 */
BFDLL BFRC BFCAPI BFDVPDeinit (void);

/* BFDVPSysBuffersAlloc -- Allocate an array of system buffers suitable for DVP acquisition. Pairs well with BiBufferAssign.
 *                         This method will expand the process working memory pool as necessary.
 *  sysBufArray     (IN)  Array of unassigned buffer pointers big enough for bufferCnt buffers
 *                  (OUT) Array of allocated buffers.
 *  width           The width of each buffer, in pixels.
 *  height          The height of each buffer, in pixels.
 *  bitdepth        The bit-depth of each buffer.
 *  bufferCnt       The number of buffers to allocate.
 *  params          Optional allocation parameters (nullptr for defaults).
 */
BFDLL BFRC BFCAPI BFDVPSysBuffersAlloc ( BFDVPSysBufArray sysBufArray, BFU32 width, BFU32 height, BFU32 bitdepth, BFU32 bufferCnt, PBFDVPAllocParams params );

/* BFDVPSysBuffersFree -- Free an array of buffers allocated by BFDVPSysBuffersAlloc. Typically called after BiBufferUnassign.
 *  sysBufArray     Array of pointers to system buffers that will be freed.
 *  bufferCnt       The number of buffers to free.
 */
BFDLL BFRC BFCAPI BFDVPSysBuffersFree ( BFDVPSysBufArray sysBufArray, BFU32 bufferCnt );

/* BFDVPBindPair -- Pair the given system and CUDA buffers; the system buffer must be allocated as pinned, aligned memory,
 *                  and should be big enough to hold the entire image. See BFDVPSysBuffersAlloc.
 *  pBufPair        A pointer to an empty buffer pairing.
 *  sysBuffer       The system buffer to be paired with the given CUDA buffer, devBuffer.
 *  devBuffer       The CUDA buffer to be paired with the given system buffer, sysBuffer.
 *  width           The width of each buffer, in pixels.
 *  height          The height of each buffer, in pixels.
 *  bitdepth        The bit-depth of each buffer.
 */
BFDLL BFRC BFCAPI BFDVPBindPair ( PBFDVPSysDevPair pBufPair, BFDVPSysBuf sysBuffer, CUdeviceptr devBuffer, BFU32 width, BFU32 height, BFU32 bitdepth );

/* BFDVPBreakPair -- Breaks the buffer pair. The caller is responsible for freeing the buffers.
 *  pBufPair        The device buffer pairing to break. No buffers will be freed.
 */
BFDLL BFRC BFCAPI BFDVPBreakPair ( PBFDVPSysDevPair pBufPair );

/* BFDVPAllocAndBindPairs -- Allocate arrays of system and CUDA buffers suitable for DVP acquisition, and automatically
 *                           bind each index-matched system and CUDA buffer pair. sysBufArray pairs well with BiBufferAssign.
 *  sysBufArray     (IN)  Array of bufferCnt unassigned system buffer pointers.
 *                  (OUT) Array of allocated system buffers.
 *  bufPairArray    (IN)  Array of bufferCnt empty system-device buffer pairings.
 *                  (OUT) Array of configured system-device buffer pairings.
 *  width           The width of each buffer, in pixels.
 *  height          The height of each buffer, in pixels.
 *  bitdepth        The bit-depth of each buffer.
 *  bufferCnt       The number of buffer pairs to allocate and make.
 *  params          Optional allocation and pairing parameters (ignored if nullptr).
 */
BFDLL BFRC BFCAPI BFDVPAllocAndBindPairs ( BFDVPSysBufArray sysBufArray, PBFDVPSysDevPair bufPairArray, BFU32 width, BFU32 height, BFU32 bitdepth, BFU32 bufferCnt, PBFDVPAllocParams params );

/* BFDVPBreakAndFreePairs -- Break pairs established by BFDVPAllocAndBindPairs, and free the system and CUDA buffers.
 *                           Typically called after BiBufferUnassign.
 *  bufPairArray    Array of system-device buffer pairings to break and free. The system and device buffers will all be freed.
 *  bufferCnt       The number of buffer pairs to break and free.
 */
BFDLL BFRC BFCAPI BFDVPBreakAndFreePairs ( PBFDVPSysDevPair bufPairArray, BFU32 bufferCnt );

/* BFDVPSys2Dev -- DMA from the system buffer to the device buffer.
 *  pBufPair        The buffer pairing on which to execute the transfer.
 *  params          Optional transfer parameters (ignored if nullptr).
 */
BFDLL BFRC BFCAPI BFDVPSys2Dev ( PBFDVPSysDevPair pBufPair, PBFDVPTransferParams params );

/* BFDVPDev2Sys -- DMA from the device buffer to the system buffer.
 *  pBufPair        The buffer pairing on which to execute the transfer.
 *  params          Optional transfer parameters (ignored if nullptr).
 */
BFDLL BFRC BFCAPI BFDVPDev2Sys ( PBFDVPSysDevPair pBufPair, PBFDVPTransferParams params );

#ifdef __cplusplus
}
#endif

#endif // INCLUDED__B_F__D_V_P__API__H
